"""
Unified HTTP handler for downloading artifacts from HTTP/HTTPS URLs.

Combines functionality of HttpSourceHandler and HttpFetchHandler
into a single handler following the Command/Query pattern.
"""

import asyncio
import hashlib
import tempfile
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Callable, Dict, Optional  # Added Dict, Any

from conduit.core.commands import (
    BaseCommand,
    BaseResult,
    CalculateMetadataCommand,
    CalculateMetadataResult,
    FetchArtifactCommand,
    FetchArtifactResult,
    ValidateTargetCommand,
    ValidateTargetResult,
)

from ..core.clients.http import AiohttpClient, HttpClient
from ..core.models import CacheMetadata
from .base import BaseHandler


class HttpHandler(BaseHandler):
    """
    Handler for HTTP/HTTPS artifacts.

    Handles:
    - Downloading to cache during lock
    - Metadata calculation with caching
    """

    def can_handle(self, scheme: str) -> bool:
        """Check if this handler can handle the given scheme."""
        return scheme in ("http", "https")

    async def handle(self, command: BaseCommand) -> BaseResult:
        """
        Route command to appropriate handler method.

        Args:
            command: Command to execute

        Returns:
            Command execution result
        """
        if isinstance(command, CalculateMetadataCommand):
            return await self._calculate_metadata(command)
        if isinstance(command, FetchArtifactCommand):
            return await self._fetch_artifact(command)
        if isinstance(command, ValidateTargetCommand):
            return await self._validate_target(command)
        msg = f"Unsupported command type: {type(command)}"
        raise ValueError(msg)

    async def _calculate_metadata(
        self, command: CalculateMetadataCommand
    ) -> CalculateMetadataResult:
        """
        Calculate metadata by downloading to cache.

        This is used during generate phase to download artifacts and calculate checksums.
        """
        try:
            if not self._cache_service:
                return CalculateMetadataResult(
                    success=False,
                    checksum="",
                    size=0,
                    type="http",
                    error_message="Cache service required for HTTP metadata calculation",
                )

            # Check if already in cache
            if await self._cache_service.is_available_locally(command.origin):
                resolution = await self._cache_service.resolve_artifact(command.origin)
                if resolution.cached_metadata:
                    content_path = self._cache_service._content_store.get_content_path(
                        resolution.cached_metadata.content_hash
                    )
                    return CalculateMetadataResult(
                        success=True,
                        checksum=resolution.cached_metadata.content_hash,
                        size=resolution.cached_metadata.size,
                        type="http",
                        cached_path=str(content_path),
                    )

            # Download to cache
            result = await self._download_to_cache(
                url=command.origin,  # Pass url explicitly
                artifact_name=command.artifact_name,
                progress_callback=command.progress_callback,  # Pass the callback
            )

            return CalculateMetadataResult(
                success=True,
                checksum=result["checksum"],
                size=result["size"],
                type="http",
                cached_path=result["cached_path"],
            )

        except Exception as e:
            return CalculateMetadataResult(
                success=False, checksum="", size=0, type="http", error_message=str(e)
            )

    async def _fetch_artifact(
        self, command: FetchArtifactCommand
    ) -> FetchArtifactResult:
        """Fetch artifact from HTTP URL to target location, always using the cache."""
        try:
            target_path = Path(command.target_path)

            if not self._cache_service:
                return FetchArtifactResult(
                    success=False,
                    local_path="",
                    checksum="",
                    was_downloaded=False,
                    error_message="Cache service is required for HTTP fetch operations.",
                )

            # Step 1: Ensure the artifact is in the cache and get its details.
            # _download_to_cache will fetch from origin if not cached or if existing cache item is invalid.
            # It needs to be aware of command.expected_checksum if provided.
            # For now, _download_to_cache doesn't explicitly use expected_checksum to avoid re-download.
            # We might need to enhance _download_to_cache or CacheResolver.is_available_locally for this.

            # Let's first check if the target already has the correct file.
            # This is an optimization to avoid even cache interaction if target is already perfect.
            if command.expected_checksum and target_path.exists():
                if await self._should_skip_download(
                    target_path, command.expected_checksum
                ):
                    # If target is already correct, we can return early.
                    return FetchArtifactResult(
                        success=True,
                        local_path=str(target_path),
                        checksum=command.expected_checksum,
                        was_downloaded=False,
                    )

            # Ensure artifact is in cache. This might trigger a download from origin.
            # The artifact_name for progress reporting could be derived or passed.
            # For simplicity, let's derive it from the URL for now.
            artifact_name_for_cache = (
                command.origin_url.split("/")[-1] or command.origin_url
            )

            cached_data = await self._download_to_cache(
                url=command.origin_url,
                artifact_name=artifact_name_for_cache,
                # We need a way for _download_to_cache to respect command.expected_checksum
                # to avoid re-downloading if a valid version is already cached.
                # This is a gap: _download_to_cache currently relies on CacheResolver.is_available_locally
                # which checks metadata, but not necessarily a specific checksum if multiple entries exist for a URL.
            )

            if not cached_data or not cached_data.get("cached_path"):
                return FetchArtifactResult(
                    success=False,
                    local_path="",
                    checksum="",
                    was_downloaded=False,
                    error_message="Failed to ensure artifact is in cache.",
                )

            cached_file_path = Path(cached_data["cached_path"])
            actual_checksum_from_cache = cached_data["checksum"]

            # If an expected_checksum was provided, verify the cached item matches.
            if (
                command.expected_checksum
                and command.expected_checksum != actual_checksum_from_cache
            ):
                return FetchArtifactResult(
                    success=False,
                    local_path="",
                    checksum=actual_checksum_from_cache,
                    was_downloaded=False,  # It was in cache, but wrong checksum
                    error_message=f"Cached artifact checksum {actual_checksum_from_cache} does not match expected {command.expected_checksum}.",
                )

            # Step 2: Copy from cache to the target_path.
            target_path.parent.mkdir(parents=True, exist_ok=True)

            # Use the cache service's content store to copy the file
            # This assumes _cache_service._content_store is ContentAddressableStore
            copied_to_target = await self._cache_service._content_store.copy_to_target(
                content_hash=actual_checksum_from_cache,  # Use the actual hash of the content in cache
                target_path=target_path,
            )

            if not copied_to_target:
                return FetchArtifactResult(
                    success=False,
                    local_path=str(target_path),
                    checksum=actual_checksum_from_cache,
                    was_downloaded=False,  # Not downloaded from origin by this command, but copy failed
                    error_message=f"Failed to copy artifact from cache {cached_file_path} to target {target_path}.",
                )

            # Determine if a "download" from origin happened during the _download_to_cache call.
            # This is tricky. _download_to_cache doesn't directly return this.
            # For now, if the file at target_path was just created/updated by copy_to_target,
            # and it wasn't already there with the right checksum, we can consider it "materialized" by this command.
            # The `was_downloaded` flag's meaning might need to be "was materialized at target by this command".

            # Let's assume if we reached here, the file is now at target_path correctly.
            # The `was_downloaded` flag should reflect if the _download_to_cache call
            # actually hit the network. This information isn't directly available from _download_to_cache's current return.
            # For simplicity now, if the file wasn't already at target_path perfectly, we'll say it was "materialized".
            # A more accurate `was_downloaded_from_origin` would require _download_to_cache to report it.

            return FetchArtifactResult(
                success=True,
                local_path=str(target_path),
                checksum=actual_checksum_from_cache,
                was_downloaded=True,  # Simplified: means file was placed at target by this op.
                # Does not strictly mean "downloaded from internet by *this* FetchArtifactCommand call"
                bytes_downloaded=cached_data["size"],  # Size of the artifact
            )

        except Exception as e:
            return FetchArtifactResult(
                success=False,
                local_path="",
                checksum="",
                was_downloaded=False,
                error_message=str(e),
            )

    async def _validate_target(self, _: ValidateTargetCommand) -> ValidateTargetResult:
        """HTTP handler doesn't handle targets, only sources."""
        return ValidateTargetResult(
            valid=False,
            error_message="HTTP handler only supports artifact sources, not targets",
        )

    async def _download_with_client(
        self, command: FetchArtifactCommand, client: HttpClient, target_path: Path
    ) -> FetchArtifactResult:
        """Download file using provided HTTP client."""
        temp_file = target_path.with_suffix(target_path.suffix + ".tmp")

        try:
            result = await client.download_file(command.origin_url, str(temp_file))

            if not result["success"]:
                return FetchArtifactResult(
                    success=False,
                    local_path="",
                    checksum="",
                    was_downloaded=False,
                    error_message=result.get("error", "Download failed"),
                )

            # Calculate checksum and move to final location
            checksum = await self._calculate_file_checksum(temp_file)
            temp_file.rename(target_path)

            return FetchArtifactResult(
                success=True,
                local_path=str(target_path),
                checksum=checksum,
                was_downloaded=True,
                bytes_downloaded=result.get("bytes_downloaded", 0),
            )

        except Exception as e:
            if temp_file.exists():
                temp_file.unlink()
            raise e

    # The _download_with_client method is now effectively replaced by _download_to_cache + copy from cache.
    # We can remove _download_with_client.
    # async def _download_with_client(...) -> FetchArtifactResult: ...

    async def _download_to_cache(
        self,
        url: str,
        artifact_name: Optional[str] = None,
        progress_callback: Optional[
            Callable[[Dict[str, Any]], None]
        ] = None,  # Updated signature
    ) -> dict:
        """Download to cache and return metadata."""
        if not self._cache_service:
            # This case should ideally be caught by the calling method (_calculate_metadata or _fetch_artifact)
            # but adding a check here for safety.
            msg = "Cache service is required for _download_to_cache operation."
            raise Exception(msg)

        with tempfile.NamedTemporaryFile(delete=False) as temp_file:
            temp_path = Path(temp_file.name)

        try:
            # Create wrapper for progress callback
            wrapped_callback = None  # type: ignore
            if progress_callback and artifact_name:

                def wrapped_callback(event):
                    event["name"] = artifact_name
                    progress_callback(event)

            # Download to temporary location
            # command = FetchArtifactCommand(origin_url=url, target_path=str(temp_path))

            client = self._http_client
            if client:
                download_result = await client.download_file(
                    url, str(temp_path), progress_callback=wrapped_callback
                )
            else:
                async with AiohttpClient() as temp_client:
                    download_result = await temp_client.download_file(
                        url, str(temp_path), progress_callback=wrapped_callback
                    )

            if not download_result["success"]:
                msg = f"Download failed: {download_result.get('error')}"
                raise Exception(msg)

            # Calculate checksum
            checksum = await self._calculate_file_checksum(temp_path)
            size = temp_path.stat().st_size

            # Store in cache
            metadata_repo = self._cache_service._metadata_repo
            content_store = self._cache_service._content_store

            await content_store.store_content(temp_path, checksum)
            content_path = content_store.get_content_path(checksum)

            # Store metadata
            cache_metadata = CacheMetadata(
                url=url,
                content_hash=checksum,
                etag=None,
                last_modified=None,
                cache_time=datetime.now(timezone.utc),
                headers={},
                size=size,
            )
            await metadata_repo.store_metadata(cache_metadata)

            return {
                "checksum": checksum,
                "size": size,
                "cached_path": str(content_path),
            }

        finally:
            if temp_path.exists():
                temp_path.unlink()

    async def _should_skip_download(
        self, target_path: Path, expected_checksum: Optional[str]
    ) -> bool:
        """Check if download should be skipped."""
        if not target_path.exists() or not expected_checksum:
            return False

        try:
            actual_checksum = await self._calculate_file_checksum(target_path)
            return actual_checksum == expected_checksum
        except Exception:
            return False

    async def _calculate_file_checksum(self, file_path: Path) -> str:
        """Calculate SHA256 checksum asynchronously."""

        def _calc():
            sha256_hash = hashlib.sha256()
            with open(file_path, "rb") as f:
                for chunk in iter(lambda: f.read(8192), b""):
                    sha256_hash.update(chunk)
            return f"sha256:{sha256_hash.hexdigest()}"

        loop = asyncio.get_event_loop()
        return await loop.run_in_executor(None, _calc)
